@@ -1,7 +1,7 @@
 ;;; w3.el --- Main functions for emacs-w3 on all platforms/versions
 ;; Author: wmperry
-;; Created: 1998/12/30 11:26:51
-;; Version: 1.6
+;; Created: 1999/11/09 14:52:37
+;; Version: 1.10
 ;; Keywords: faces, help, comm, news, mail, processes, mouse, hypermedia
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -83,6 +83,7 @@
 (require 'cl)
 (require 'css)
 (require 'url-vars)
+(require 'url-parse)
 (require 'w3-vars)
 (eval-and-compile
   (require 'w3-display))
@@ -143,16 +144,18 @@ the subprocess exits."
 
 (defun w3-viewer-filter (proc string)
   ;; A process filter for asynchronous external viewers
-  (let ((buff (get-buffer-create (url-generate-new-buffer-name
-				  (symbol-name
-				   (read (nth 2 (process-command proc))))))))
-    (save-excursion
-      (set-buffer buff)
-      (erase-buffer)
-      (insert string)
-      (set-process-buffer proc buff)
-      (w3-notify-when-ready buff)
-      (set-process-filter proc nil))))
+  (if (= (length string) 0)
+      nil
+    (let ((buff (get-buffer-create (url-generate-new-buffer-name
+				    (symbol-name
+				     (read (nth 2 (process-command proc))))))))
+      (save-excursion
+	(set-buffer buff)
+	(erase-buffer)
+	(insert string)
+	(set-process-buffer proc buff)
+	(w3-notify-when-ready buff)
+	(set-process-filter proc nil)))))
 
 (defun w3-viewer-sentinel (proc string)
   ;; Delete any temp files left from a viewer process.
@@ -410,7 +413,7 @@ With prefix argument, use the URL of the hyperlink under point instead."
 	   (string-match url-nonrelative-link (car command-line-args-left)))
       (setq url (car command-line-args-left)
 	    command-line-args-left (cdr command-line-args-left)))
-  (if (equal url "") (error "No document specified!"))
+  (if (or (null url) (equal url "")) (error "No document specified!"))
   ;; legal use for relative URLs ?
   (if (string-match "^www:[^/].*" url)
       (setq url (concat (file-name-directory (url-filename
@@ -1097,14 +1100,6 @@ insert URL under point"
 	(setq x (url-buffer-visiting "www:/auto/hotlist"))))
     (w3-fetch "www://auto/hotlist")))
 
-(defun url-maybe-relative (url)
-  "Take a url and either fetch it, or resolve relative refs, then fetch it"
-  (cond
-   ((not
-     (string-match url-nonrelative-link url))
-    (w3-relative-link url))
-   (t (w3-fetch url))))
-
 (defun w3-in-assoc (elt list)
   "Check to see if ELT matches any of the regexps in the car elements of LIST"
   (let (rslt)
@@ -1185,6 +1180,8 @@ invokes some commands which read a coding system from the user.")
 	       (;; explicit charset ? (through MIME headers or META tag)
 		(and (stringp mmcharset)
 		     (setq coding-system (w3-coding-system-for-mime-charset mmcharset))))
+	       (;; recorded explicit coding system ? (in `w3-explicit-encodings-file')
+		(setq coding-system (w3-recall-explicit-coding-system url-current-object)))
 	       ;; forced conversion ? (through user option w3-force-conversion-alist)
 	       ((and (url-filename url-current-object)
 		     (let ((force-alist w3-force-conversion-alist)
@@ -1236,6 +1233,7 @@ invokes some commands which read a coding system from the user.")
 	  (if (or (stringp proc)
 		  (bufferp proc)) (setq w3-current-last-buffer proc))
 	  (remove-hook 'after-change-functions 'url-after-change-function)
+	  (if (fboundp 'clear-progress) (clear-progress))
 	  (if url-be-asynchronous
 	      (progn
 		(cond
@@ -1781,9 +1779,97 @@ With prefix argument, it reads a coding system to decode the document."
 	(w3-explicit-coding-system explicit-coding-system))
     (kill-buffer (current-buffer))
     (w3-fetch tmp)
+    (if explicit-coding-system
+	(w3-record-explicit-coding-system tmp explicit-coding-system))
     (goto-char pnt)
     (set-window-start (selected-window) (min window-start (point-max)))))
 
+(defun w3-recall-explicit-coding-system (url)
+  "Find user-specified explicit coding system for this URL
+in w3-explicit-conversion-tree"
+  (let* ((urlobj (if (stringp url)
+		     (url-generic-parse-url url)
+		   url))
+	 (hostname (or (url-host urlobj) "localhost"))
+	 (fname-list (split-string (url-filename urlobj) "\\/")))
+    ;; now recurse
+    (w3-find-explicit-coding-system (cons hostname fname-list) w3-explicit-conversion-tree)))
+
+(defun w3-find-explicit-coding-system (fname-list tree)
+  "Recall a user-specified explicit coding system"
+  (let ((branch (assoc (car fname-list) tree)))
+    (and branch
+	 (or (and (cdr fname-list) (cddr branch)
+		  (w3-find-explicit-coding-system (cdr fname-list) (cddr branch)))
+	     (cadr branch)))))
+
+(defun w3-record-explicit-coding-system (url coding-system)
+  "Record user-specified explicit coding system for URLs
+as high as possible in w3-explicit-conversion-tree"
+  (let* ((urlobj (if (stringp url)
+		     (url-generic-parse-url url)
+		   url))
+	 (hostname (or (url-host urlobj) "localhost"))
+	 (fname-list (split-string (url-filename urlobj) "\\/"))
+	 (tree (or (assoc hostname w3-explicit-conversion-tree)
+		   (let ((branch (list hostname)))
+		     (setq w3-explicit-conversion-tree
+			   (cons branch w3-explicit-conversion-tree))
+		     branch))))
+    ;; now recurse
+    (w3-add-explicit-coding-system fname-list coding-system tree)
+    (setq w3-explicit-encodings-changed-since-last-save t)))
+
+(defun w3-add-explicit-coding-system (fname-list coding-system tree)
+  "Memorize a user-specified explicit coding system"
+  (if (and (cadr tree) (not (equal (cadr tree) coding-system)))
+      (setcar (cdr tree) nil))
+  (let ((branch (assoc (car fname-list) (cddr tree))))
+    (cond (branch
+	   ;; update existing branch
+	   (cond ((cdr fname-list)
+		  (or (equal (cadr branch) coding-system)
+		      (null (cadr branch))
+		      (setcar (cdr branch) nil))
+		  (w3-add-explicit-coding-system (cdr fname-list) coding-system branch))
+		 (t
+		  (setcar (cdr branch) coding-system))))
+	  (t
+	   ;; create a new branch
+	   (setcdr tree
+		   (if fname-list
+		       (let ((subbranch (list (car fname-list))))
+			 (w3-add-explicit-coding-system (cdr fname-list) coding-system subbranch)
+			 (cons (if (or (null (cddr tree))
+				       (equal coding-system (cadr tree)))
+				   coding-system)
+			       (cons subbranch (cddr tree))))
+		     (list coding-system)))))))
+
+(defun w3-write-explicit-encodings (&optional fname)
+  "Write the explicit encodings file into `w3-explicit-encodings-file'."
+  (interactive)
+  (or fname
+      (and w3-explicit-encodings-file
+	   (setq fname (expand-file-name w3-explicit-encodings-file))))
+  (cond
+   ((not w3-explicit-encodings-changed-since-last-save) nil)
+   ((not (file-writable-p fname))
+    (message "Explicit encodings file %s (see variable `w3-explicit-encodings-file') is unwritable." fname))
+   (t
+    (let ((make-backup-files nil)
+	  (version-control nil)
+	  (require-final-newline t))
+      (save-excursion
+	(set-buffer (get-buffer-create " *w3-tmp*"))
+	(erase-buffer)
+	(insert "(setq w3-explicit-conversion-tree\n      '"
+		(prin1-to-string w3-explicit-conversion-tree)
+		")\n\n")
+	(write-file fname)
+	(kill-buffer (current-buffer))))))
+  (setq w3-explicit-encodings-changed-since-last-save nil))
+
 (defun w3-leave-buffer ()
   "Bury this buffer, but don't kill it."
   (interactive)
@@ -1928,7 +2014,7 @@ No arg means whole window full.  Arg is number of lines to scroll."
 	(found nil))
     (setq found (cdr-safe (assoc "reply-to" url-current-mime-headers)))
     (if (and found (not (string-match url-nonrelative-link found)))
-	(setq found (list (concat "mailto:" found))))
+	(setq found (list (list 'href (concat "mailto:" found)))))
     (while (and x (not found))
       (setq y (car x)
 	    x (cdr x)
@@ -2175,6 +2261,13 @@ dumped with emacs."
 			       "Please consult the `%s' buffer for details."))
 			    w3-default-configuration-file buf-name))))))
 	       
+  ;; Load the explicit encodings file if it exists
+  (if (and w3-explicit-encodings-file
+	   (file-exists-p w3-explicit-encodings-file))
+      (condition-case nil
+	  (load w3-explicit-encodings-file nil t)
+	(error nil)))
+
   (if (and (eq w3-user-colors-take-precedence 'guess)
 	   (not (eq (device-type) 'tty))
 	   (not (eq (device-class) 'mono)))
